import cv2
import imutils
import numpy as np
import matplotlib.pyplot as plt
from skimage.util import img_as_ubyte
from random import Random
import random
from skimage.transform import SimilarityTransform
from skimage.transform import warp
from skimage.feature import hog


def translation_image(image, r, tx_min, tx_max, ty_min, ty_max, show_info=False):
    """ Realitza una translació de la imatge.
    El vector de translació tindrà com a component x un valor
    aleatori entre tx_min i tx_max. Com a component y tindrà un valor aleatori entre ty_min i ty_max.
    Segueix els passos indicats a https://docs.opencv.org/master/da/d6e/tutorial_py_geometric_transformations.html

    Args:
        :param image: Imatge a la que aplicarem la translació.
        :param r: Llavor per a número aleatoris.
        :param tx_min: Del vecslació,tor de tran el valor x mínim que pot agafar.
        :param tx_max: Del vector de translació, el valor x màxim que pot agafar.
        :param ty_min: Del vector de translació, el valor y mínim que pot agafar.
        :param ty_max: Del vector de translació, el valor y màxim que pot agafar.
        :param show_info: Booleà que ens indica si mostrarem informació. Defaults False (boolean, optional).

    Returns:
        :return image: Retorna la imatge amb la translació.
    """
    rows, cols = image.shape
    tx = r.randint(tx_min, tx_max)
    ty = r.randint(ty_min, ty_max)
    matrix = np.float32([[1, 0, tx], [0, 1, ty]])

    if show_info:
        print("translation random modification tx: ", tx, ", ty: ", ty)

    return cv2.warpAffine(image, matrix, (cols, rows))


def rotate_image(image, degrees_min, degrees_max, r, show_info=False):
    """ Transforma la imatge d'entrada fent la rotació indicada als paràmetres.
    L'àngle de rotació serà un número aleatori entre degrees_min i degrees_max.
    Segueix els passos indicats a https://www.pyimagesearch.com/2017/01/02/rotate-images-correctly-with-opencv-and-python/

    Args:
        :param image: Imatge d'entrada.
        :param degrees_min: Mínim número de graus.
        :param degrees_max: Màxim número de graus.
        :param r: Llavor per a número aleatoris.
        :param show_info: Booleà que ens indica si mostrarem informació. Defaults False (boolean, optional).

    Returns:
        :return image: Retorna la imatge amb la rotació.
    """
    graus = r.randint(degrees_min, degrees_max)
    if show_info:
        print("degrees rotation random modification: ", graus)

    return imutils.rotate_bound(image, graus)


def gamma_image(image, gamma_min, gamma_max, show_info=False):
    """ Modifica el valor gamma de la imatge d'entrada. El valor gamma serà aleatori entre gamma_min i gamma_max.
    S'ha seguit el mètode indicat a https://www.pyimagesearch.com/2015/10/05/opencv-gamma-correction/

    Args:
        :param image: Imatge d'entrada
        :param gamma_min: Valor mínim de gamma.
        :param gamma_max: Valor màxim de gamma.
        :param show_info: show_info: Booleà que ens indica si mostrarem informació. Defaults False (boolean, optional).

    Returns:
        :return image: Retorna la imatge amb el canvi de gamma.
    """
    gamma = random.uniform(gamma_min, gamma_max)
    inv_gamma = 1.0 / gamma

    if show_info:
        print("gamma random modification: ", gamma)

    table = np.array([((i / 255.0) ** inv_gamma) * 255
                      for i in np.arange(0, 256)]).astype("uint8")

    return cv2.LUT(image, table)


def rand_warp_image(image, r=Random(),
                    tp=10, tx_min=-10, tx_max=10, ty_min=-10, ty_max=10,
                    degrees_p=10, degrees_min=0, degrees_max=360,
                    gamma_p=10, gamma_min=0.1, gamma_max=2.0,
                    show_info=False):
    """ Modifica la imatge que rep com a paràmetre.
    Pot modificar-se aplicant: traslacions, rotació, aumentació o canvi d'iluminació.
    Com a paràmetres tenim la imatge, els valors màxims que aplicarem a cada transformació i
    el % per saber si apliquem o no una determinada transformació.

    Args:
        :param image: Imatge que modificarem
        :param r: Llavor per a número aleatoris. Defaults Random()(Random, optional)
        :param tp: Indica percentatge de realiztar la transformació de translació. Contra més alt aquest valor és MÉS probable. Defaults 10 (int, optional).
        :param tx_min: Del vector de translació, el valor x mínim que pot agafar. Defaults -10 (int, optional).
        :param tx_max: Del vector de translació, el valor x màxim que pot agafar. Defaults 10 (int, optional).
        :param ty_min: Del vector de translació, el valor y mínim que pot agafar. Defaults -10 (int, optional).
        :param ty_max: Del vector de translació, el valor y màxim que pot agafar. Defaults 10 (int, optional).
        :param degrees_p: Indica percentatge de realiztar la transformació de rotació. Contra més alt aquest valor és MÉS probable. Defaults 10 (int, optional).
        :param degrees_min: Mínim número de graus. Defaults 0 (int, optional).
        :param degrees_max: Màxim número de graus. Defaults 360 (int, optional).
        :param gamma_p: Indica percentatge de realiztar la transformació gamma. Contra més alt aquest valor és MENYS probable. Defaults 10 (int, optional).
        :param gamma_min: Valor mínim de gamma. Defaults 0.1 (float, optional).
        :param gamma_max: Valor màxim de gamma. Defaults 2.0 (float, optional).
        :param show_info: Booleà que ens indica si mostrarem informació. Defaults True (boolean, optional).

    Returns:
        :return image: Imatge amb la modificació.
    """
    switch_flag = r.randint(0, tp)
    if switch_flag:
        image = translation_image(image, r, tx_min, tx_max, ty_min, ty_max, show_info=show_info)

    switch_flag = r.randint(0, degrees_p)
    if switch_flag:
        image = rotate_image(image, degrees_min, degrees_max, r, show_info=show_info)

    switch_flag = r.randint(0, gamma_p)
    if not switch_flag:
        image = gamma_image(image, gamma_min, gamma_max, show_info=show_info)

    return image


def generate_pair_images(skimage, r, w=320, h=240, show_info=False):
    """ Crea un parell d'imatges per a l'entrenament donada la imatge original.

    Donada la imatge original es procedeix a:
        -1- Es converteix a escala de grisos.
        -2- Es redimensiona a w*h.
        -3- Es crea una nova imatge canviada.
        -4- D'una de les dues imatges, escollida de forma aleatoria, es calcula HOG.
        -5- Es retorna la imatge i els descriptors HOG.

    Args:
        :param skimage: Imatge original en format skimage.
        :param r: Llavor per a la generació de número aleatòris.
        :param w: Amplada de la imatge transformada. Defaults 320 (int, optional).
        :param h: Alçada de la imatge transformada. Defaults 240 (int, optional).
        :param show_info: Indica si es mostrarà les imatges generades al final de l'execució. Defaults fals (boolean, optional).

    Returns:
        :return image_processed: imatge processada
        :return hog_vector: descriptors HOG
    """

    image = img_as_ubyte(skimage)

    if show_info:
        image_original = image

    if len(image.shape) > 2:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    # FIXME: en este punto se hace una u otra transformacion
    image_warp = rand_warp_image(image, show_info=show_info)
    #image_warp = random_transform(image)

    image = cv2.resize(image, (w, h), interpolation=cv2.INTER_CUBIC)
    image_warp = cv2.resize(image_warp, (w, h), interpolation=cv2.INTER_CUBIC)

    # La llavor s'ha d'inicialitzar correctament o sempre retornarà el mateix valor.
    switch_flag = r.randint(0, 1)
    if switch_flag:
        image_processed = image_warp
        hog_vector = hog(image, orientations=8, pixels_per_cell=(10, 10), cells_per_block=(1, 1), multichannel=False)
    else:
        image_processed = image
        hog_vector = hog(image_warp, orientations=8, pixels_per_cell=(10, 10), cells_per_block=(1, 1),
                         multichannel=False)

    if show_info:
        is_warped = "No warped image"
        if switch_flag:
            is_warped = "Warped image"

        fig, (ax1, ax2, ax3) = plt.subplots(1, 3, figsize=(8, 4), sharex=False, sharey=False)

        ax1.axis('off')
        ax1.imshow(image_original)
        ax1.set_title('Input image')

        ax2.axis('off')
        ax2.imshow(image_processed)
        ax2.set_title('Processed image:\n ' + is_warped)

        ax3.hist(hog_vector)
        ax3.set_title("HOG")

        plt.show()

    return image_processed, hog_vector


def image_transform(image, scale_factor, rotation_angle, tx_factor, ty_factor):
    center_y, center_x = np.array(image.shape[:2]) / 2.
    rotation = SimilarityTransform(rotation=np.deg2rad(rotation_angle))
    zoom = SimilarityTransform(scale=scale_factor)
    shift = SimilarityTransform(translation=[-center_x, -center_y])
    shift_inv = SimilarityTransform(translation=[center_x, center_y])
    translation = SimilarityTransform(translation=[tx_factor * 2 * center_x, ty_factor * 2 * center_y])

    return warp(image, (shift + (rotation + shift_inv)) + (shift + (zoom + shift_inv)) + translation, mode='reflect')


def random_transform(image, scale_min=0.9, scale_max=1.1, rotate_min=-90, rotate_max=90, tx_min=-0.25, tx_max=0.25,
                     ty_min=-0.25, ty_max=0.25):
    scale_factor = np.random.uniform(scale_min, scale_max)
    rotation_angle = np.random.uniform(rotate_min, rotate_max)
    tx_factor = np.random.uniform(tx_min, tx_max)
    ty_factor = np.random.uniform(ty_min, ty_max)

    return image_transform(image, scale_factor, rotation_angle, tx_factor, ty_factor)
