import keras
from keras.models import Sequential
from keras.layers import Dense, Conv2D, Flatten, MaxPooling2D, BatchNormalization, ZeroPadding2D, SpatialDropout2D
from model_utils import euclidean_distance_loss
from datetime import datetime
from swish import Swish


class Model:
    """ Estructura de la xarxa neuronal. """

    def __init__(self, w=320, h=240, filtres_sortida=6144, name=None, show_info=False):
        """ Model de la xarxa neuronal.

            Args:
                :param w: Amplada a la que es redimensionarà la imatge. Defaults 320 (int, optional).
                :param h: Alçada a la que es redimensionarà la imatge. Defaults 240 (int, optional).
                :param filtres_sortida: Número de filtres que tindrà la sortida. Defaults 6144 (int, optional).
                :param name: Nom del model. Serveix per desar el model. Si no es defineix, es crearà un de propi. Defaults None (str, optional).
                :param show_info: Serveix per mostrar informació durant el procés. Defaults False (bool, optional).
        """
        self.name = name
        self.__generate_name__()
        self.w = w
        self.h = h
        self.filtres_sortida = filtres_sortida
        self.model = Sequential()

        # INICI CODIFICACIO:
        # 1er conjunt de capes
        # Amb aquest zero padding safegeix el padding a Keras:
        self.model.add(ZeroPadding2D(padding=(4, 4), input_shape=(h, w, 1)))
        # random_uniform equival a la distribució gausiana
        self.model.add(
            Conv2D(64, kernel_size=5, strides=2, padding='valid',
                   kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                   bias_initializer=keras.initializers.RandomNormal(stddev=0.01), activation='swish'))
        self.model.add(MaxPooling2D(pool_size=3, strides=2))
        self.model.add(BatchNormalization())

        # 2on conjunt de capes
        self.model.add(ZeroPadding2D(padding=2))
        self.model.add(
            Conv2D(128, kernel_size=4, strides=1, padding='valid',
                   kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                   bias_initializer=keras.initializers.RandomNormal(stddev=0.01), activation='swish'))
        self.model.add(MaxPooling2D(pool_size=3, strides=2))
        self.model.add(BatchNormalization())

        # 3er conjunt de capes
        self.model.add(
            Conv2D(4, kernel_size=3, strides=1, padding='valid',
                   kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                   bias_initializer=keras.initializers.RandomNormal(stddev=0.01), activation='swish'))

        self.model.add(Flatten())
        # FI CODIFICACIO ---

        # INICI DESCODIFICACIO:
        # 1ra capa completament connectada
        self.model.add(
            Dense(1064, activation='sigmoid', kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                  bias_initializer=keras.initializers.RandomNormal(stddev=0.01)))

        # 2na capa completament connectada
        self.model.add(
            Dense(1024, activation='sigmoid', kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                  bias_initializer=keras.initializers.RandomNormal(stddev=0.01)))

        # 3ra capa completament connectada
        self.model.add(
            Dense(filtres_sortida, activation='sigmoid',
                  kernel_initializer=keras.initializers.RandomNormal(stddev=0.01),
                  bias_initializer=keras.initializers.RandomNormal(stddev=0.01)))
        # FI DESCODIFICACIO---

        self.model.compile(
            optimizer='sgd',
            loss=euclidean_distance_loss,
            metrics=['mae']
        )

        if show_info:
            self.model.summary()

    def __generate_name__(self):
        """ Generació del nom del model.
            Si el model no té nom, genera un de nom. Mètode privat.
        """
        if not self.name:
            self.name = "MODEL_" + datetime.now().strftime("%Y%m%d-%H%M%S")
