from pickle import dump, load
from keras.models import load_model
from model_utils import euclidean_distance_loss
import matplotlib.pyplot as plt
from neural_network_utils import build_filenames


class NeuralNetwork:
    """ Classe de la xarxa neuronal que ens proporciona un conjunt de mètodes per poder operar amb aquesta."""

    def __init__(self, model=None):
        if model:
            self.model = model.model
            self.name = model.name
            self.shape = self.model.input_shape[1:4]
        self.history = None

    def train_neural_network(self, train_generator, val_generator=None, epochs=10):
        """ Mètode per entrenar la xarxa neuronal.
            Actualitzarà la història de la xarxa.

            Args:
                :param train_generator: Dades d'entrenament. De tipus DataGenerator.
                :param epochs: Número de èpoques que entrenarà la xarxa neuronal. Defaults 10 (int, optional).
                :param val_generator: Dades de validació.  De tipus DataGenerator.
        """
        self.history = self.model.fit_generator(train_generator, epochs=epochs, validation_data=val_generator).history

    def evaluate(self, test_generator):
        """ Evalua la xarxa neuronal.

            Args:
                :param test_generator: Dades per avaluar el model. De tipus DataGenerator.
        """
        print("metricas: ", self.model.metrics_names)
        print(self.model.evaluate_generator(test_generator))

    def predict(self, entrada):
        """ Realitza una predicció.

            Args:
                :param entrada: Entrada de la que realitzarà la predicció.
        """
        return self.model.predict(entrada)

    def save(self, filename=None):
        """ Desa el model.

            Args:
                :param filename: Nom amb el que es desarà el model. Si no rep cap, se li assignarà el del model. Defaults None (str, optional).
        """
        if not filename:
            filename = self.name
        model_name, hist_name = build_filenames(filename)
        self.model.save(model_name)
        with open(hist_name, 'wb') as hist_file:
            dump(self.history, hist_file)

    def load(self, filename):
        """ Carrega un model indicat.

            Args:
                :param filename: Nom del model que es carregarà.
        """
        model_name, hist_name = build_filenames(filename)
        self.model = load_model(model_name, custom_objects={'euclidean_distance_loss': euclidean_distance_loss})
        self.name = filename
        self.shape = self.model.input_shape[1:4]
        with open(hist_name, 'rb') as hist_file:
            self.history = load(hist_file)

    def plot_training_history(self, plot_title='TRAINING EVOLUTION', show_info=False):
        """ Dibuixa el gràfic d'entrenament.

            Args:
                :param plot_title: Nom del gràfic.
                :param show_info: Mostrarà informació detallada a la consola.
        """
        if show_info:
            print(self.history.keys())

        plt.plot(self.history['loss'])
        plt.plot(self.history['val_loss'])

        plt.title(plot_title)
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'validation'], loc='upper left')
        plt.show()

        plt.title("Error Absolut Mitjà")
        plt.plot(self.history['mean_absolute_error'])
        plt.ylabel('mae')
        plt.xlabel('epoch')
        plt.legend(['mae'], loc='upper left')
        plt.show()
