import numpy as np
from skimage.transform import resize
from imagematcher import ImageMatcher
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
import time
import random


class NeuralNetworkEval:
    """ Classe per avaluar una xarxa neuronal. """

    def __init__(self, neural_network, test_data_set, batch_size=5):
        self.neural_network = neural_network
        self.test_data_set = test_data_set
        self.batch_size = batch_size
        self.hit_ratio = None
        self.auc = None
        self.stats = None
        self.times = None
        self.matcher = ImageMatcher()

    def predict_images(self, number_images, img_input):
        """ Realitza prediccions sobre un conjunt d'imatges.

            Args:
                :param number_images: Número d'imatges que s'avaluaràn.
                :param img_input: Conjunt d'imatges que s'avaluaràn.

            Returns:
                :return: Conjunt de prediccions de les imatges d'entrada.
        """
        input_shape = self.neural_network.shape
        number_batches = int(np.ceil(float(number_images) / float(self.batch_size)))

        model = self.neural_network.model
        predictions = np.zeros((number_images, model.layers[-1].output_shape[1]))

        for batch_iteration in range(number_batches):
            min_iteration = batch_iteration * self.batch_size
            max_iteration = min((batch_iteration + 1) * self.batch_size, number_images)
            images = np.zeros((max_iteration - min_iteration, input_shape[0], input_shape[1], input_shape[2]))
            for image_iteration in range(min_iteration, max_iteration):
                images[image_iteration - min_iteration, ] = resize(img_input(image_iteration), input_shape)
                predictions[min_iteration:max_iteration, ] = self.neural_network.predict(images)

        return predictions

    def compute_hitratio_evolution(self, distance_type='euclidean'):
        """ Analitza el percentatge d'encert.

            Args:
                :param distance_type: Distància que s'emprarà per calcular el grau d'encert. Defaults euclidean (str, optional).

            Returns:
                :return: Percentatge d'encert.
                :return: Mitjana d'encert.
        """
        db_pred = self.predict_images(self.test_data_set.numDBImages, self.test_data_set.get_dbimage)
        q_pred = self.predict_images(self.test_data_set.numQImages, self.test_data_set.get_qimage)
        distances = cdist(q_pred, db_pred, distance_type)
        self.hit_ratio = []

        for pct in range(100):
            number_items = int(pct * self.test_data_set.numDBImages / 100)
            self.hit_ratio.append(self.__compute_hitratio__(distances, number_items))
        self.auc = np.sum(self.hit_ratio) / 10000

        return self.hit_ratio, self.auc

    def __compute_hitratio__(self, distances, number_items):
        number_found = 0
        for query_index in range(self.test_data_set.numQImages):
            db_loop_candidates = np.argsort(distances[query_index, :])[:number_items]
            db_actual_loops = self.test_data_set.get_qloop(query_index)
            found_loop = (any(i in db_actual_loops for i in db_loop_candidates))
            if found_loop:
                number_found += 1
        hit_ratio = number_found * 100 / self.test_data_set.numQImages

        return hit_ratio

    def plot_hit_ratio_evolution(self):
        """ Dibuixa una gràfica amb el percentatge d'encert. """
        if not self.hit_ratio or not self.auc:
            self.compute_hitratio_evolution()
        plt.figure()
        plt.plot(self.hit_ratio)
        plt.title('AUC=' + str(int(self.auc * 100)))
        plt.ylabel('Hit ratio (%)')
        plt.xlabel('Percentage of database')
        plt.show()

    def compute_fullstats(self, search_ratio=0.1, distance_type='euclidean'):
        """ Calcula la matriu de confusió. """
        print('[COMPUTING FULLSTATS]')
        print('  * COMPUTING DESCRIPTORS')
        time_start = time.time()
        db_prediction = self.predict_images(self.test_data_set.numDBImages, self.test_data_set.get_dbimage)
        q_prediction = self.predict_images(self.test_data_set.numQImages, self.test_data_set.get_qimage)
        print('  * COMPUTING DISTANCES BETWEEN DESCRIPTORS')
        distances = cdist(q_prediction, db_prediction, distance_type)
        time_end = time.time() - time_start
        number_items = int(search_ratio * self.test_data_set.numDBImages)
        print('  * THE CLOSEST ' + str(number_items) + ' DATABASE IMAGES WILL BE SEARCHED')
        print('  * STARTING TESTS')
        tp = tn = fp = fn = 0
        time_start = time.time()
        for query_index in range(self.test_data_set.numQImages):
            db_loop_candidates = np.argsort(distances[query_index, :])[:number_items]
            query_image = self.test_data_set.get_qimage(query_index)
            db_actual_loops = self.test_data_set.get_qloop(query_index)
            tn_num = 0
            for database_index in range(self.test_data_set.numDBImages):
                is_loop = database_index in db_actual_loops
                found_loop = False
                if database_index in db_loop_candidates:
                    database_image = self.test_data_set.get_dbimage(database_index)
                    self.matcher.define_images(query_image, database_image)
                    found_loop = self.matcher.estimate()
                    del database_image
                if found_loop and is_loop:
                    tp += 1
                    print("**tp** query_index:" + str(query_index) + ", database_index: " + str(database_index))
                elif found_loop and (not is_loop):
                    fp += 1
                    print("**fp** query_index:" + str(query_index) + ", database_index: " + str(database_index))
                elif (not found_loop) and is_loop:
                    fn += 1
                    print("**fn** query_index:" + str(query_index) + ", database_index: " + str(database_index))
                elif (not found_loop) and (not is_loop):
                    tn += 1
                    tn_num = tn_num + 1
                    if tn_num < 3:
                        print("**tn** query_index:" + str(query_index) + ", database_index: " + str(database_index))
            print('    + COMPLETED ' + str(query_index + 1) + ' OF ' + str(self.test_data_set.numQImages) + ' QUERIES')
            del query_image
        time_loops = time.time() - time_start
        print('  * TESTS FINISHED')
        self.stats = (tp, fp, tn, fn)
        self.times = (time_end, time_loops)
        try:
            accuracy = (tp + tn) / (tp + tn + fp + fn)
        except ZeroDivisionError:
            accuracy = 0
        try:
            tpr = tp / (tp + fn)
        except ZeroDivisionError:
            tpr = 0
        try:
            fpr = fp / (fp + tn)
        except ZeroDivisionError:
            fpr = 0
        try:
            prec = tp / (tp + fp)
        except ZeroDivisionError:
            prec = 0
        try:
            f1 = 2 * ((tpr * prec) / (tpr + prec))
        except ZeroDivisionError:
            f1 = 0

        print('[FULLSTATS COMPUTED accuracy:' + str(accuracy) + ', tpr: ' + str(tpr) + ', fpr: ' + str(fpr)
              + ', prec: ' + str(prec) + ', f1: ' + str(f1) + ', time: ' + str(time_loops + time_end)
              + ', tp: ' + str(tp) + ', fp: ' + str(fp) + ', tn: ' + str(tn) + ', fn: ' + str(fn)
              + ', time_end: ' + str(time_end) + ', time_loops: ' + str(time_loops) + ']')

        return tp, fp, tn, fn, time_end, time_loops

    def check_one_query(self, query_index=16):
        """Agafa una imatge aleatoria que tanca un bucle i troba les més properes al conjunt d'estudi.

            Args:
                :param query_index: Índex de la imatge de tancament de bucle que s'emprarà per a la prova. Defaults 16 (int, optional).

            Returns:
                :return: Retorna si la xarxa neuronal ha detectat un bucle real donat.
        """

        db_prediction = self.predict_images(self.test_data_set.numDBImages, self.test_data_set.get_dbimage)
        query_prediction = self.predict_images(self.test_data_set.numQImages, self.test_data_set.get_qimage)
        #query_index = random.randrange(0, self.test_data_set.numQImages-1)

        distances = cdist(query_prediction, db_prediction)
        db_loop_candidates = np.argsort(distances[query_index, :])[:self.test_data_set.numDBImages]
        db_actual_loops = self.test_data_set.get_qloop(query_index)

        #np.isin(db_actual_loops, db_loop_candidates)
        return np.where(db_loop_candidates == db_actual_loops)
